@@ -39,9 +39,17 @@ For more details, see individual docstrings.
 
 from __future__ import annotations
 from typing import TYPE_CHECKING
-
-if TYPE_CHECKING:
-    from typing import Tuple, List, Dict, Optional, Any, TypedDict, Union
+from typing import (
+    Tuple,
+    List,
+    Dict,
+    Optional,
+    Any,
+    TypedDict,
+    Union,
+    TypeVar,
+    Type,
+)
 
 from itertools import zip_longest
 from copy import deepcopy
@@ -50,6 +58,7 @@ from enum import Enum
 from attr import define, field, Factory, cmp_using, setters  # type: ignore
 import numpy as np
 import numpy.typing as npt
+from scipy.sparse import coo_array
 from numpy.lib import recfunctions as rfn
 from typicle import Types
 from typicle.convert import cast_array
@@ -58,14 +67,21 @@ from ._base import ParticleBase, AdjacencyBase, MaskBase, ArrayBase
 from .calculate import eta, phi
 
 
-def _is_np_structured(array: np.ndarray) -> bool:
-    return array.dtype.names is not None
-
-
 ###########################################
 # SET UP ARRAY ATTRIBUTES FOR DATACLASSES #
 ###########################################
 _types = Types()
+DoubleVector = npt.NDArray[np.float64]
+BoolVector = npt.NDArray[np.bool_]
+IntVector = npt.NDArray[np.int32]
+HalfIntVector = npt.NDArray[np.int16]
+ObjVector = npt.NDArray[np.object_]
+AnyVector = npt.NDArray[Any]
+DataType = TypeVar("DataType", bound=ArrayBase)
+
+
+def _is_np_structured(array: AnyVector) -> bool:
+    return array.dtype.names is not None
 
 
 def array_field(type_name: str):
@@ -75,9 +91,9 @@ def array_field(type_name: str):
     default = Factory(lambda: np.array([], dtype=dtype))
     equality_comparison = cmp_using(np.array_equal)
 
-    def converter(values: npt.ArrayLike) -> np.ndarray:
+    def converter(values: npt.ArrayLike) -> AnyVector:
         data = np.array(values)
-        out_array: np.ndarray = cast_array(data, cast_type=dtype)
+        out_array: AnyVector = cast_array(data, cast_type=dtype)
         return out_array
 
     return field(
@@ -101,7 +117,7 @@ class MaskAggOp(Enum):
 class MaskArray(MaskBase, ArrayBase):
     """Data structure for containing masks over particle data."""
 
-    data: np.ndarray = array_field("bool")
+    data: BoolVector = array_field("bool")
 
     def copy(self) -> "MaskArray":
         return deepcopy(self)
@@ -114,10 +130,10 @@ class MaskArray(MaskBase, ArrayBase):
     def __len__(self) -> int:
         return len(self.data)
 
-    def __array__(self) -> np.ndarray:
+    def __array__(self) -> BoolVector:
         return self.data
 
-    def __and__(self, other: Union[MaskBase, np.ndarray]) -> MaskArray:
+    def __and__(self, other: Union[MaskBase, BoolVector]) -> MaskArray:
         if isinstance(other, MaskBase):
             other_data = other.data
         elif isinstance(other, np.ndarray):
@@ -129,7 +145,7 @@ class MaskArray(MaskBase, ArrayBase):
             )
         return self.__class__(np.bitwise_and(self.data, other_data))
 
-    def __or__(self, other: Union[MaskBase, np.ndarray]) -> MaskArray:
+    def __or__(self, other: Union[MaskBase, BoolVector]) -> MaskArray:
         if isinstance(other, MaskBase):
             other_data = other.data
         elif isinstance(other, np.ndarray):
@@ -146,7 +162,7 @@ class MaskArray(MaskBase, ArrayBase):
 
 
 if TYPE_CHECKING:
-    _IN_MASK_DICT = Dict[str, Union[MaskArray, np.ndarray]]
+    _IN_MASK_DICT = Dict[str, Union[MaskArray, BoolVector]]
     _MASK_DICT = Dict[str, MaskArray]
 
 
@@ -215,14 +231,17 @@ class MaskGroup(MaskBase):
             mask = MaskArray(mask)
         self._mask_arrays.update({key: mask})
 
+    def __len__(self) -> int:
+        return len(self.data)
+
     def __delitem__(self, key) -> None:
         """Remove a MaskArray from the group, using given key."""
         self._mask_arrays.pop(key)
 
-    def __array__(self) -> np.ndarray:
+    def __array__(self) -> BoolVector:
         return self.data
 
-    def __and__(self, other: Union[MaskBase, np.ndarray]) -> MaskArray:
+    def __and__(self, other: Union[MaskBase, BoolVector]) -> MaskArray:
         if isinstance(other, MaskBase):
             other_data = other.data
         elif isinstance(other, np.ndarray):
@@ -234,7 +253,7 @@ class MaskGroup(MaskBase):
             )
         return MaskArray(np.bitwise_and(self.data, other_data))
 
-    def __or__(self, other: Union[MaskBase, np.ndarray]) -> MaskArray:
+    def __or__(self, other: Union[MaskBase, BoolVector]) -> MaskArray:
         if isinstance(other, MaskBase):
             other_data = other.data
         elif isinstance(other, np.ndarray):
@@ -257,19 +276,19 @@ class MaskGroup(MaskBase):
         return list(self._mask_arrays.keys())
 
     @property
-    def bitwise_or(self) -> np.ndarray:
+    def bitwise_or(self) -> BoolVector:
         return np.bitwise_or.reduce(  # type: ignore
             [child.data for child in self._mask_arrays.values()]
         )
 
     @property
-    def bitwise_and(self) -> np.ndarray:
+    def bitwise_and(self) -> BoolVector:
         return np.bitwise_and.reduce(  # type: ignore
             [child.data for child in self._mask_arrays.values()]
         )
 
     @property
-    def data(self) -> np.ndarray:
+    def data(self) -> BoolVector:
         """Same as MaskGroup.bitwise_and."""
         if self.agg_op is MaskAggOp.AND:
             return self.bitwise_and
@@ -286,7 +305,7 @@ class MaskGroup(MaskBase):
             )
 
     @property
-    def dict(self) -> Dict[str, np.ndarray]:
+    def dict(self) -> Dict[str, BoolVector]:
         return {key: val.data for key, val in self._mask_arrays.items()}
 
 
@@ -326,7 +345,7 @@ class PdgArray(ArrayBase):
 
     from mcpid.lookup import PdgRecords as __PdgRecords
 
-    data: np.ndarray = array_field("int")
+    data: IntVector = array_field("int")
     __lookup_table: __PdgRecords = field(init=False, repr=False)
     __mega_to_giga: float = field(init=False, repr=False)
 
@@ -337,7 +356,7 @@ class PdgArray(ArrayBase):
     def __len__(self) -> int:
         return len(self.data)
 
-    def __array__(self) -> np.ndarray:
+    def __array__(self) -> IntVector:
         return self.data
 
     def __getitem__(self, key) -> "PdgArray":
@@ -385,65 +404,65 @@ class PdgArray(ArrayBase):
             np.isin(data, target, assume_unique=False, invert=blacklist)
         )
 
-    def __get_prop(self, field: str) -> np.ndarray:
+    def __get_prop(self, field: str) -> AnyVector:
         props = self.__lookup_table.properties(self.data, [field])[field]
         return props  # type: ignore
 
-    def __get_prop_range(self, field: str) -> np.ndarray:
+    def __get_prop_range(self, field: str) -> AnyVector:
         field_range = [field + "lower", field + "upper"]
         props = self.__lookup_table.properties(self.data, field_range)
         return props  # type: ignore
 
     @property
-    def name(self) -> np.ndarray:
+    def name(self) -> ObjVector:
         return self.__get_prop("name")
 
     @property
-    def charge(self) -> np.ndarray:
+    def charge(self) -> DoubleVector:
         return self.__get_prop("charge")
 
     @property
-    def mass(self) -> np.ndarray:
-        out: np.ndarray = self.__get_prop("mass") * self.__mega_to_giga
+    def mass(self) -> DoubleVector:
+        out: DoubleVector = self.__get_prop("mass") * self.__mega_to_giga
         return out
 
     @property
-    def mass_bounds(self) -> np.ndarray:
+    def mass_bounds(self) -> AnyVector:
         range_arr = self.__get_prop_range("mass")
         for name in range_arr.dtype.names:
             range_arr[name] *= self.__mega_to_giga
         return range_arr
 
     @property
-    def quarks(self) -> np.ndarray:
+    def quarks(self) -> ObjVector:
         return self.__get_prop("quarks")
 
     @property
-    def width(self) -> np.ndarray:
-        out: np.ndarray = self.__get_prop("width") * self.__mega_to_giga
+    def width(self) -> DoubleVector:
+        out: DoubleVector = self.__get_prop("width") * self.__mega_to_giga
         return out
 
     @property
-    def width_bounds(self) -> np.ndarray:
+    def width_bounds(self) -> AnyVector:
         range_arr = self.__get_prop_range("width")
         for name in range_arr.dtype.names:
             range_arr[name] *= self.__mega_to_giga
         return range_arr
 
     @property
-    def isospin(self) -> np.ndarray:
+    def isospin(self) -> DoubleVector:
         return self.__get_prop("i")
 
     @property
-    def g_parity(self) -> np.ndarray:
+    def g_parity(self) -> DoubleVector:
         return self.__get_prop("g")
 
     @property
-    def space_parity(self) -> np.ndarray:
+    def space_parity(self) -> DoubleVector:
         return self.__get_prop("p")
 
     @property
-    def charge_parity(self) -> np.ndarray:
+    def charge_parity(self) -> DoubleVector:
         return self.__get_prop("c")
 
 
@@ -467,12 +486,12 @@ class MomentumArray(ArrayBase):
         Azimuthal component of particle momenta.
     """
 
-    data: np.ndarray = array_field("pmu")
+    data: AnyVector = array_field("pmu")
 
     def __len__(self) -> int:
         return len(self.data)
 
-    def __array__(self) -> np.ndarray:
+    def __array__(self) -> AnyVector:
         return self.data
 
     def __getitem__(self, key) -> "MomentumArray":
@@ -493,35 +512,35 @@ class MomentumArray(ArrayBase):
         return vec
 
     @property
-    def pt(self) -> npt.NDArray[np.float64]:
+    def pt(self) -> DoubleVector:
         """Momentum component transverse to the beam-axis."""
         return self._vector.pt  # type: ignore
 
     @property
-    def eta(self) -> npt.NDArray[np.float64]:
+    def eta(self) -> DoubleVector:
         """Pseudorapidity of particles."""
         return eta(self)
 
     @property
-    def phi(self) -> npt.NDArray[np.float64]:
+    def phi(self) -> DoubleVector:
         """Azimuthal angular displacement of particles about the
         beam-axis.
         """
         return phi(self)
 
     @property
-    def theta(self) -> npt.NDArray[np.float64]:
+    def theta(self) -> DoubleVector:
         """Spherical angular displacement of particles from the positive
         beam-axis.
         """
         return self._vector.theta  # type: ignore
 
     @property
-    def mass(self) -> npt.NDArray[np.float64]:
+    def mass(self) -> DoubleVector:
         """Mass of the particles."""
         return self._vector.mass  # type: ignore
 
-    def delta_R(self, other_pmu: "MomentumArray") -> npt.NDArray[np.float64]:
+    def delta_R(self, other_pmu: "MomentumArray") -> DoubleVector:
         return self._vector.deltaR(other_pmu._vector)  # type: ignore
 
 
@@ -539,7 +558,7 @@ class ColorArray(ArrayBase):
         Structured array containing color / anti-color pairs.
     """
 
-    data: np.ndarray = array_field("color")
+    data: AnyVector = array_field("color")
 
     def copy(self):
         return deepcopy(self)
@@ -552,7 +571,7 @@ class ColorArray(ArrayBase):
     def __len__(self) -> int:
         return len(self.data)
 
-    def __array__(self) -> np.ndarray:
+    def __array__(self) -> AnyVector:
         return self.data
 
 
@@ -570,7 +589,7 @@ class HelicityArray(ArrayBase):
         Helicity values.
     """
 
-    data: np.ndarray = array_field("helicity")
+    data: HalfIntVector = array_field("helicity")
 
     def copy(self) -> "HelicityArray":
         """Returns a new StatusArray instance with same data."""
@@ -584,7 +603,7 @@ class HelicityArray(ArrayBase):
     def __len__(self) -> int:
         return len(self.data)
 
-    def __array__(self) -> np.ndarray:
+    def __array__(self) -> HalfIntVector:
         return self.data
 
 
@@ -606,7 +625,7 @@ class StatusArray(ArrayBase):
     produced the data.
     """
 
-    data: np.ndarray = array_field("h_int")
+    data: HalfIntVector = array_field("h_int")
 
     def copy(self) -> "StatusArray":
         """Returns a new StatusArray instance with same data."""
@@ -620,7 +639,7 @@ class StatusArray(ArrayBase):
     def __len__(self) -> int:
         return len(self.data)
 
-    def __array__(self) -> np.ndarray:
+    def __array__(self) -> HalfIntVector:
         return self.data
 
     def in_range(
@@ -729,12 +748,12 @@ class ParticleSet(ParticleBase):
     @classmethod
     def from_numpy(
         cls,
-        pdg: Optional[np.ndarray] = None,
-        pmu: Optional[np.ndarray] = None,
-        color: Optional[np.ndarray] = None,
-        helicity: Optional[np.ndarray] = None,
-        status: Optional[np.ndarray] = None,
-        final: Optional[np.ndarray] = None,
+        pdg: Optional[IntVector] = None,
+        pmu: Optional[AnyVector] = None,
+        color: Optional[AnyVector] = None,
+        helicity: Optional[HalfIntVector] = None,
+        status: Optional[HalfIntVector] = None,
+        final: Optional[BoolVector] = None,
     ) -> "ParticleSet":
         """Creates a ParticleSet instance directly from numpy arrays.
 
@@ -762,7 +781,9 @@ class ParticleSet(ParticleBase):
             objects, and providing a unified interface to them.
         """
 
-        def optional(data_class, data: Optional[np.ndarray]):
+        def optional(
+            data_class: Type[DataType], data: Optional[AnyVector]
+        ) -> DataType:
             return data_class(data) if data is not None else data_class()
 
         return cls(
@@ -802,13 +823,13 @@ class AdjacencyList(AdjacencyBase):
         Adjacency matrix representation.
     """
 
-    _data: np.ndarray = array_field("edge")
-    weights: np.ndarray = array_field("double")
+    _data: AnyVector = array_field("edge")
+    weights: DoubleVector = array_field("double")
 
     def __len__(self) -> int:
         return len(self._data)
 
-    def __array__(self) -> np.ndarray:
+    def __array__(self) -> AnyVector:
         return self._data
 
     def __getitem__(self, key) -> "AdjacencyList":
@@ -843,7 +864,10 @@ class AdjacencyList(AdjacencyBase):
 
     @classmethod
     def from_matrix(
-        cls, adj_matrix: np.ndarray, weighted: bool = False
+        cls,
+        adj_matrix: Union[DoubleVector, BoolVector, IntVector],
+        weighted: bool = False,
+        self_loop: bool = False,
     ) -> "AdjacencyList":
         """Construct an AdjacencyList object from an optionally weighted
         adjacency matrix.
@@ -857,15 +881,20 @@ class AdjacencyList(AdjacencyBase):
             Whether or not to propogate the numerical values in the
             elements of the adjacency matrix as the edge weights
             (default: False).
+        self_loop : bool
+            If True will add edges from nodes to themselves, with weight
+            0 (if applicable). (default: True)
         """
-        edges = np.where(adj_matrix)
-        kwargs = {"data": np.array(edges).T}
+        sps_adj = coo_array(adj_matrix)
+        if self_loop is True:
+            sps_adj.setdiag(0.0)
+        kwargs = {"data": np.vstack((sps_adj.row, sps_adj.col)).T}
         if weighted is True:
-            kwargs["weights"] = adj_matrix[edges]
+            kwargs["weights"] = sps_adj.data
         return cls(**kwargs)
 
     @property
-    def matrix(self) -> np.ndarray:
+    def matrix(self) -> Union[DoubleVector, IntVector]:
         size = len(self.nodes)
         if len(self.weights) > 0:
             weights = self.weights
@@ -878,11 +907,11 @@ class AdjacencyList(AdjacencyBase):
         return adj
 
     @property
-    def edges(self) -> np.ndarray:
+    def edges(self) -> AnyVector:
         return self._data
 
     @property
-    def nodes(self) -> np.ndarray:
+    def nodes(self) -> IntVector:
         """Nodes are extracted from the edge list, and put in
         ascending order of magnitude, regardless of sign.
         Positive sign conventionally means final state particle.
@@ -895,8 +924,8 @@ class AdjacencyList(AdjacencyBase):
 
     def to_dicts(
         self,
-        edge_data: Optional[Dict[str, Union[ArrayBase, np.ndarray]]] = None,
-        node_data: Optional[Dict[str, Union[ArrayBase, np.ndarray]]] = None,
+        edge_data: Optional[Dict[str, Union[ArrayBase, AnyVector]]] = None,
+        node_data: Optional[Dict[str, Union[ArrayBase, AnyVector]]] = None,
     ) -> _AdjDict:
         """Returns data in dictionary format, which is more easily
         parsed by external libraries, such as NetworkX.
@@ -908,10 +937,10 @@ class AdjacencyList(AdjacencyBase):
 
         def make_data_dicts(
             orig: Tuple[Any, ...],
-            data: Dict[str, Union[ArrayBase, np.ndarray]],
+            data: Dict[str, Union[ArrayBase, AnyVector]],
         ):
             def array_iterator(array_dict):
-                for array in data.values():
+                for array in array_dict.values():
                     if isinstance(array, ArrayBase):
                         yield array.data
                     elif isinstance(array, np.ndarray):
@@ -988,7 +1017,8 @@ class Graphicle:
         kwargs = dict()
         for name in self.__attr_names:
             data = getattr(self, name)
-            kwargs.update({name: data[key]})
+            if len(data) != 0:
+                kwargs.update({name: data[key]})
         return self.__class__(**kwargs)
 
     def copy(self) -> "Graphicle":
@@ -997,14 +1027,14 @@ class Graphicle:
     @classmethod
     def from_numpy(
         cls,
-        pdg: Optional[np.ndarray] = None,
-        pmu: Optional[np.ndarray] = None,
-        color: Optional[np.ndarray] = None,
-        helicity: Optional[np.ndarray] = None,
-        status: Optional[np.ndarray] = None,
-        final: Optional[np.ndarray] = None,
-        edges: Optional[np.ndarray] = None,
-        weights: Optional[np.ndarray] = None,
+        pdg: Optional[IntVector] = None,
+        pmu: Optional[AnyVector] = None,
+        color: Optional[AnyVector] = None,
+        helicity: Optional[HalfIntVector] = None,
+        status: Optional[HalfIntVector] = None,
+        final: Optional[BoolVector] = None,
+        edges: Optional[AnyVector] = None,
+        weights: Optional[DoubleVector] = None,
     ) -> "Graphicle":
         """Instantiates a Graphicle object from an optional collection
         of numpy arrays.
@@ -1081,11 +1111,11 @@ class Graphicle:
         return self.particles.final
 
     @property
-    def edges(self) -> np.ndarray:
+    def edges(self) -> AnyVector:
         return self.adj.edges
 
     @property
-    def nodes(self) -> np.ndarray:
+    def nodes(self) -> IntVector:
         return self.adj.nodes
 
     def _need_attr(self, attr_name: str, task: str) -> None:
